//==========================================
// Matt Pietrek
// Originally from October 2000 MSDN Magazine
// Modified in July 2001
//
// **** YOU MUST MODIFY THE PROJECT'S "Additional include directories" TO
// **** POINT AT THE SDK'S INCLUDE DIRECTORY
//==========================================

#include "stdafx.h"
#include "MetaDataImportHelper.h"

#include "CorError.h"

void DisplayHeaderInfo( IMetaDataImport * pIMetaData, LPCSTR pszFileName );
void DisplayAssemblyInfo( IMetaDataAssemblyImport * pIMetaDataAssembly );
void EnumAssemblyRefs( IMetaDataAssemblyImport * pIMetaDataAssembly );
void DisplayTypeDefAttrs( CorTypeAttr dwTypeDefAttr );
void DisplayTypeDefMember( IMetaDataImport * pIMetaData, mdToken memberToken );
void EnumTypeDefMembers( IMetaDataImport * pIMetaData, mdTypeDef td );
void EnumTypeDefProps( IMetaDataImport * pIMetaData, mdTypeDef td );
void DisplayTypeDef( IMetaDataImport * pIMetaData, mdTypeDef td );
void EnumTypeDefs( IMetaDataImport * pIMetaData );
void DisplayTypeRef( IMetaDataImport * pIMetaData, mdTypeRef tr );
void EnumTypeRefs( IMetaDataImport * pIMetaData );
void EnumMemberRefs( IMetaDataImport * pIMetaData, mdToken refToken );
void EnumModuleRefs( IMetaDataImport * pIMetaData );
void FormatGUID( GUID & guid, char * pszOut );

//=========================================================

int main(int argc, char* argv[])
{
    CoInitializeEx( NULL, COINIT_APARTMENTTHREADED );

    if ( argc < 2 )
    {
        printf( "Syntax: Meta <filename>\n" );
        return 1;
    }

    CMetaDataImportHelper * pMetaDataHelper = 0;
    try
    {
        // New a CMetaDataImportHelper class.  This lets us
        // catch the exception if the metadata isn't loaded
        // properly by the constructor
        pMetaDataHelper = new CMetaDataImportHelper( argv[1] );
    }
    catch( char * psz )
    {
        printf( psz );
        return 1;
    }

    IMetaDataImport * pIMetaData = pMetaDataHelper->m_pIMetaDataImport;

    DisplayHeaderInfo( pIMetaData, argv[1] );

    printf( "=================================================\n" );

    DisplayAssemblyInfo( pMetaDataHelper->m_pIMetaDataAssemblyImport );

    printf( "=================================================\n" );
    
    EnumAssemblyRefs( pMetaDataHelper->m_pIMetaDataAssemblyImport );

    printf( "=================================================\n" );

    EnumTypeDefs( pIMetaData );

    printf( "=================================================\n" );
    
    EnumTypeRefs( pIMetaData );
    
    printf( "=================================================\n" );
    
    EnumModuleRefs( pIMetaData );
    
    delete pMetaDataHelper;

    CoUninitialize();

    return 0;
}

//=========================================================
void DisplayHeaderInfo( IMetaDataImport * pIMetaData, LPCSTR pszFileName )
{
    wchar_t wszScopeName[512];
    ULONG cchScopeName = ARRAYSIZE(wszScopeName);
    GUID mvid;

    pIMetaData->GetScopeProps( wszScopeName, cchScopeName, &cchScopeName, &mvid );
    
    char szMVID[128];
    FormatGUID( mvid, szMVID );
    
    printf( "File:      %s\n", pszFileName );
    printf( "ScopeName: %ls\n", wszScopeName );
    printf( "MVID:      %s\n", szMVID );
}

//=========================================================

void DisplayAssemblyInfo( IMetaDataAssemblyImport * pIMetaDataAssembly )
{
    wchar_t wszName[512];
    ULONG   cchName = ARRAYSIZE(wszName);
    wchar_t wszTitle[512];
    ULONG   cchTitle = ARRAYSIZE(wszTitle);
    wchar_t wszDescription[512];
    ULONG   cchDescription = ARRAYSIZE(wszDescription);
    DWORD   dwAssemblyFlags;

    HRESULT hr =
    pIMetaDataAssembly->GetAssemblyProps(
			(DWORD)mdtAssembly +1,  // [IN] The Assembly for which to get the properties.(HACK!)
			0,						// [OUT] Pointer to the public key.
			0,						// [OUT] Count of bytes in the public key.
			0,						// [OUT] Hash Algorithm.
			wszName,			    // [OUT] Buffer to fill with name.
			cchName,				// [IN] Size of buffer in wide chars.
			&cchName,				// [OUT] Actual # of wide chars in name.
			0,						// [OUT] Assembly MetaData.
            &dwAssemblyFlags);		// [OUT] Flags.

    if ( FAILED(hr) )
        return;

    printf( "Assembly: %ls\n", wszName );
    printf( "Title: %ls\n", wszTitle );
    printf( "Description: %ls\n", wszDescription );
}

//=========================================================

void EnumAssemblyRefs( IMetaDataAssemblyImport * pIMetaDataAssembly )
{
    printf( "Assembly Refs\n" );

    HCORENUM hEnum = 0;
    mdAssemblyRef   rAssemblyRefs[1024];
    ULONG           cTokens = ARRAYSIZE(rAssemblyRefs);

    HRESULT hr =
    pIMetaDataAssembly->EnumAssemblyRefs(
                            &hEnum,             // [IN|OUT] Pointer to the enum.
                            rAssemblyRefs,      // [OUT] Put AssemblyRefs here.
                            cTokens,            // [IN] Max AssemblyRefs to put.
                            &cTokens);          // [OUT] Put # put here.

    if ( FAILED(hr) )
        return;

    for ( ULONG i = 0; i < cTokens; i++ )
    {
        ASSEMBLYMETADATA assemblyMetaData;
        wchar_t wszName[512];
        ULONG   cchName = ARRAYSIZE(wszName);
		DWORD dwAssemblyRefFlags;

        memset( &assemblyMetaData, 0, sizeof(assemblyMetaData) );

        pIMetaDataAssembly->GetAssemblyRefProps(
					rAssemblyRefs[i],       // [IN] The AssemblyRef for which to get the properties.
					0,						// [OUT] Pointer to the public key or token.
					0,						// [OUT] Count of bytes in the public key or token.
					wszName,                // [OUT] Buffer to fill with name.
					cchName,                // [IN] Size of buffer in wide chars.
					&cchName,               // [OUT] Actual # of wide chars in name.
					&assemblyMetaData,      // [OUT] Assembly MetaData.
					0,						// [OUT] Hash blob.
					0,						// [OUT] Count of bytes in the hash blob.
					&dwAssemblyRefFlags);	// [OUT] Flags.


        printf( "%s: %08X %ls Flags: %08X\n",
                CMetaDataImportHelper::TokenTypeName(rAssemblyRefs[i]),
                rAssemblyRefs[i],
                wszName,
                dwAssemblyRefFlags );
    }

}

//=========================================================

void DisplayTypeDefMember( IMetaDataImport * pIMetaData, mdToken memberToken )
{
    printf( "  %s: %08X", CMetaDataImportHelper::TokenTypeName( memberToken), memberToken );

    wchar_t wszMember[512];
    ULONG   cchMember = ARRAYSIZE(wszMember);
    DWORD dwAttr;
    DWORD dwCodeRVA=0xFFFFFFFF;
    PCCOR_SIGNATURE pSig;
    ULONG cbSig;

    HRESULT hr =
    pIMetaData->GetMemberProps(  
                    memberToken,    // The member for which to get props.   
                    0,              // Put member's class here. 
                    wszMember,      // Put member's name here.  
                    cchMember,      // Size of szMember buffer in wide chars.   
                    &cchMember,     // Put actual size here 
                    &dwAttr,        // Put flags here.  
                    &pSig,          // [OUT] PCCOR_SIGNATURE * point to the blob value of meta data   
                    &cbSig,         // [OUT] ULONG * actual size of signature blob  
                    &dwCodeRVA,     // [OUT] ULONG * codeRVA    
                    0,              // [OUT] DWORD * Impl. Flags    
                    0,              // [OUT] DWORD * flag for value type. selected ELEMENT_TYPE_*   
                    0,              // [OUT] void const ** constant value 
                    0);             // [OUT] ULONG * size of constant value

    if ( FAILED(hr) )
        return;

    printf( " %ls\n", wszMember );

    // Display the signature as a series of hex bytes
    // See CorCallingConvention, CorElementType, etc... in CORHDR.H
    printf( "    sig: " );

    for ( ULONG i=0; i < cbSig; i++ )
        printf( "%02X ", *pSig++ );

    printf( "\n" );
}

//=========================================================

// Helper macros that make picking out and displaying flags
// less tedious

#define DISPLAY_FLAG( value, flag ) \
    if ( (value & flag) == flag )   \
        printf( "%s ", #flag );

#define DISPLAY_MASKED_VALUE( source, value, mask ) \
    if ( (source & mask) == value )                 \
        printf( "%s ", #value );

void DisplayTypeDefAttrs( CorTypeAttr dwTypeDefAttr )
{
    DISPLAY_MASKED_VALUE( dwTypeDefAttr, tdNotPublic, tdVisibilityMask )
    DISPLAY_MASKED_VALUE( dwTypeDefAttr, tdPublic, tdVisibilityMask )
    DISPLAY_MASKED_VALUE( dwTypeDefAttr, tdNestedPublic, tdVisibilityMask )
    DISPLAY_MASKED_VALUE( dwTypeDefAttr, tdNestedPrivate, tdVisibilityMask )
    DISPLAY_MASKED_VALUE( dwTypeDefAttr, tdNestedFamily, tdVisibilityMask )
    DISPLAY_MASKED_VALUE( dwTypeDefAttr, tdNestedAssembly, tdVisibilityMask )
    DISPLAY_MASKED_VALUE( dwTypeDefAttr, tdNestedFamANDAssem, tdVisibilityMask )
    DISPLAY_MASKED_VALUE( dwTypeDefAttr, tdNestedFamORAssem, tdVisibilityMask )

    // Use this mask to retrieve class layout information
    DISPLAY_MASKED_VALUE( dwTypeDefAttr, tdAutoLayout, tdLayoutMask )
    DISPLAY_MASKED_VALUE( dwTypeDefAttr, tdSequentialLayout, tdLayoutMask )
    DISPLAY_MASKED_VALUE( dwTypeDefAttr, tdExplicitLayout, tdLayoutMask )
    // end layout mask

    // Use this mask to retrieve class semantics information.
    DISPLAY_MASKED_VALUE( dwTypeDefAttr, tdClass, tdClassSemanticsMask )
    DISPLAY_MASKED_VALUE( dwTypeDefAttr, tdInterface, tdClassSemanticsMask )
    // end semantics mask

    // Special semantics in addition to class semantics.
    DISPLAY_FLAG( dwTypeDefAttr, tdAbstract )
    DISPLAY_FLAG( dwTypeDefAttr, tdSealed )
    DISPLAY_FLAG( dwTypeDefAttr, tdSpecialName )

    // Implementation attributes.
    DISPLAY_FLAG( dwTypeDefAttr, tdImport )
    DISPLAY_FLAG( dwTypeDefAttr, tdSerializable )

    // Use tdStringFormatMask to retrieve string information for native interop
    DISPLAY_MASKED_VALUE( dwTypeDefAttr, tdAnsiClass, tdStringFormatMask )
    DISPLAY_MASKED_VALUE( dwTypeDefAttr, tdUnicodeClass, tdStringFormatMask )
    DISPLAY_MASKED_VALUE( dwTypeDefAttr, tdAutoClass, tdStringFormatMask )
    // end string format mask

    DISPLAY_FLAG( dwTypeDefAttr, tdBeforeFieldInit )

    // Flags reserved for runtime use.
    DISPLAY_MASKED_VALUE( dwTypeDefAttr, tdRTSpecialName, tdReservedMask )
    DISPLAY_MASKED_VALUE( dwTypeDefAttr, tdHasSecurity, tdReservedMask )
}

//=========================================================

void EnumTypeDefMembers( IMetaDataImport * pIMetaData, mdTypeDef td )
{
    HCORENUM    hEnum = 0;
    mdToken     rMembers[4096];
    ULONG       cTokens = ARRAYSIZE(rMembers);

    HRESULT hr = 
    pIMetaData->EnumMembers(
                    &hEnum,         // [IN|OUT] Pointer to the enum.    
                    td,             // [IN] TypeDef to scope the enumeration.   
                    rMembers,       // [OUT] Put MemberDefs here.   
                    cTokens,        // [IN] Max MemberDefs to put.  
                    &cTokens);      // [OUT] Put # put here.    

    if ( FAILED(hr) )
        return;
    
    pIMetaData->CloseEnum( hEnum ); // We don't need this HCORENUM anymore

    for ( ULONG i = 0; i < cTokens; i++ )
    {
        DisplayTypeDefMember( pIMetaData, rMembers[i] );
        printf( "\n" );
    }
}

//=========================================================

void EnumTypeDefProps( IMetaDataImport * pIMetaData, mdTypeDef td )
{
    //
    // Check for properties
    //
    mdProperty      rProperties[1024];
    ULONG           cProperties = ARRAYSIZE(rProperties);

    HCORENUM hEnum = 0;
    
    HRESULT hr = pIMetaData->EnumProperties(  
                        &hEnum,             // [IN|OUT] Pointer to the enum.    
                        td,                 // [IN] TypeDef to scope the enumeration.   
                        rProperties,        // [OUT] Put Properties here.   
                        cProperties,        // [IN] Max properties to put.  
                        &cProperties);      // [OUT] Put # put here.    

    if ( SUCCEEDED(hr) )
    {
        pIMetaData->CloseEnum( hEnum );
        
        for ( ULONG i = 0; i < cProperties; i++ )
        {
            wchar_t     wszProperty[512];
            ULONG       cchProperty = ARRAYSIZE(wszProperty);
            PCCOR_SIGNATURE pSig;
            ULONG cbSig;
            mdMethodDef mdGetter, mdSetter;

            pIMetaData->GetPropertyProps(
                    rProperties[i],     // [IN] property token  
                    0,                  // [OUT] typedef containing the property declarion. 
                    wszProperty,        // [OUT] Property name  
                    cchProperty,        // [IN] the count of wchar of szProperty    
                    &cchProperty,       // [OUT] actual count of wchar for property name    
                    0,                  // [OUT] property flags.    
                    &pSig,              // [OUT] property type. pointing to meta data internal blob 
                    &cbSig,             // [OUT] count of bytes in *ppvSig  
                    0,                  // [OUT] flag for value type. selected ELEMENT_TYPE_*   
                    0,                  // [OUT] constant value 
                    0,                  // [OUT] size of constant string in chars, 0 for non-strings.
                    &mdSetter,          // [OUT] setter method of the property  
                    &mdGetter,          // [OUT] getter method of the property  
                    0,                  // [OUT] other method of the property   
                    0,                  // [IN] size of rmdOtherMethod  
                    0);                 // [OUT] total number of other method of this property  

            printf( "  property: %08X %ls  Setter: %08X  Getter: %08X\n",
                    rProperties[i], wszProperty, mdSetter, mdGetter );

            // Display the signature as a series of hex bytes
            // See CorCallingConvention, CorElementType, etc... in CORHDR.H
            // (For methods, the first 3 bytes are interpreted as:
            //  Byte 0: Calling convention
            //  Byte 1: Number of parameters
            //  Byte 2: Return type (may be more than one byte!)
            printf( "    sig: " );

            for ( ULONG i=0; i < cbSig; i++ )
                printf( "%02X ", *pSig++ );

            printf( "\n\n" );
        }
    }

    //
    // Check for events.  I would put this code elsewhere
    // if any events actually showed up!
    //
    mdEvent     rEvents[1024];
    ULONG       cEvents = ARRAYSIZE(rEvents);

    hEnum = 0;
    hr =pIMetaData->EnumEvents(  
        &hEnum,         // [IN|OUT] Pointer to the enum.    
        td,             // [IN] TypeDef to scope the enumeration.   
        rEvents,        // [OUT] Put Events here.   
        cEvents,        // [IN] Max events to put.  
        &cEvents);      // [OUT] Put # put here.    

    if ( SUCCEEDED(hr) )
    {
        pIMetaData->CloseEnum( hEnum );
        
        if ( cEvents )
        {
            for ( ULONG i = 0; i < cEvents; i++ )
                printf( "  event: %08X\n", rEvents[i] );

            // Given an mdEvent, you can call IMetaDataImport::GetEventProps
            // to get more info on it.
        }
    }
}

//=========================================================

void DisplayTypeDef( IMetaDataImport * pIMetaData, mdTypeDef td )
{
    wchar_t wszTypeDef[512];
    ULONG   cchTypeDef = ARRAYSIZE(wszTypeDef);
    DWORD   dwTypeDefAttr;
    mdToken tkExtends;

    HRESULT hr =
    pIMetaData->GetTypeDefProps(    td,                 // [IN] TypeDef token for inquiry.
                                    wszTypeDef,         // [OUT] Put name here.
                                    cchTypeDef,         // [IN] size of name buffer in wide chars.
                                    &cchTypeDef,        // [OUT] put size of name (wide chars) here.
                                    &dwTypeDefAttr,     // [OUT] Put flags here.
                                    &tkExtends );       // [OUT] Put base class TypeDef/TypeRef here.

    if ( FAILED(hr) )
        return;

    printf( "TypeDef: %08X %ls extends %08X\n", td, wszTypeDef, tkExtends );
    
    DisplayTypeDefAttrs( (CorTypeAttr)dwTypeDefAttr );

    printf( "\n\n" );

    EnumTypeDefMembers( pIMetaData, td );

    EnumTypeDefProps( pIMetaData, td );
}

//=========================================================

void EnumTypeDefs( IMetaDataImport * pIMetaData )
{
    HCORENUM    hEnum = 0;
    mdTypeDef   rTypeDefs[2048];
    ULONG       cTypeDefs = ARRAYSIZE(rTypeDefs);
                                                            // Get "global" TypeDef
    DisplayTypeDef( pIMetaData, (mdTypeDef)0x02000001 );    // Fix this!!!
    
    HRESULT hr = pIMetaData->EnumTypeDefs( &hEnum, rTypeDefs, cTypeDefs, &cTypeDefs );
    if ( FAILED(hr) )
        return;

    pIMetaData->CloseEnum( hEnum ); // We don't need the HCORENUM open anymore

    // Spin through and process all the typedefs
    for ( ULONG i = 0; i < cTypeDefs; i++ )
    {
        printf( "=================================================\n" );
        DisplayTypeDef( pIMetaData, rTypeDefs[i] ); 
    }
}

//=========================================================

void DisplayTypeRef( IMetaDataImport * pIMetaData, mdTypeRef tr )
{
    wchar_t wszTypeRef[512];
    ULONG   cchTypeRef = ARRAYSIZE(wszTypeRef);
    mdToken tkResolutionScope;

    HRESULT hr =
    pIMetaData->GetTypeRefProps(    tr,                 // [IN] TypeRef token.
                                    &tkResolutionScope, // [OUT] Resolution scope, ModuleRef or AssemblyRef.
                                    wszTypeRef,         // [OUT] Put name here.
                                    cchTypeRef,         // [IN] size of name buffer in wide chars.
                                    &cchTypeRef);       // [OUT] put size of name (wide chars) here.

    if ( FAILED(hr) )
        return;

    printf( "TypeRef: %08X %ls  Resolution: %08X\n",
            tr, wszTypeRef, tkResolutionScope );

    // Resolve the TypeRef to an IID_IMetaDataImport for the imported assembly
    IMetaDataImport * pIMetaDataImportResolved;
    mdTypeDef typeDef;

    hr = 
    pIMetaData->ResolveTypeRef( tr,
                                IID_IMetaDataImport,
                                (LPUNKNOWN *)&pIMetaDataImportResolved,
                                &typeDef );
    GUID mvid;

    if ( SUCCEEDED(hr) )
    {
        // Get the MVID (module version ID), which is essentially a GUID
        // representing a particular build of an assembly.  Format
        // and print out the MVID (as a GUID), along with the TypeDef token
        // of the class in the imported assembly.
        pIMetaDataImportResolved->GetScopeProps( 0, 0, 0, &mvid );

        char szMVID[128];
        FormatGUID( mvid, szMVID );

        printf( "  MVID: %s\n", szMVID );

        printf( "  TypeDef: %08X\n",typeDef );
        pIMetaDataImportResolved->Release();
    }
    else
    {
        printf( "Unable to resolve TypeRef: %X\n", hr );
    }

    EnumMemberRefs( pIMetaData, tr );
    printf( "\n" );
}

//=========================================================

void EnumTypeRefs( IMetaDataImport * pIMetaData )
{
    printf( "TypeRefs\n" );

    HCORENUM    hEnum = 0;
    mdTypeDef   rTypeRefs[2048];
    ULONG       cTypeRefs = ARRAYSIZE(rTypeRefs);
    
    HRESULT hr = pIMetaData->EnumTypeRefs(  &hEnum,
                                            rTypeRefs,
                                            cTypeRefs,
                                            &cTypeRefs );
    if ( FAILED(hr) )
        return;

    pIMetaData->CloseEnum( hEnum ); // We don't need the HCORENUM open anymore

    // Spin through and process all the typeRefs
    for ( ULONG i = 0; i < cTypeRefs; i++ )
    {
        printf( "======================\n" );
        DisplayTypeRef( pIMetaData, rTypeRefs[i] ); 
    }
}

//=========================================================

void EnumMemberRefs( IMetaDataImport * pIMetaData, mdToken refToken )
{
    HCORENUM    hEnum = 0;
    mdMemberRef rMemberRefs[2048];
    ULONG       cMemberRefs = ARRAYSIZE(rMemberRefs);
    
    HRESULT hr = pIMetaData->EnumMemberRefs(&hEnum,
                                            refToken,
                                            rMemberRefs,
                                            cMemberRefs,
                                            &cMemberRefs );
    if ( FAILED(hr) )
        return;

    pIMetaData->CloseEnum( hEnum ); // We don't need the HCORENUM open anymore

    // Spin through and process all the MemberRefs
    for ( ULONG i = 0; i < cMemberRefs; i++ )
    {
        wchar_t wszMember[512];
        ULONG   cchMember = ARRAYSIZE(wszMember);

        pIMetaData->GetMemberRefProps(
                        rMemberRefs[i],     // [IN] given memberref 
                        0,                  // [OUT] Put classref or classdef here. 
                        wszMember,          // [OUT] buffer to fill for member's name   
                        cchMember,          // [IN] the count of char of szMember   
                        &cchMember,         // [OUT] actual count of char in member name    
                        0,                  // [OUT] point to meta data blob value  
                        0);                 // [OUT] actual size of signature blob  

        printf( "  %s: %08X %ls\n",
                CMetaDataImportHelper::TokenTypeName(rMemberRefs[i]),
                rMemberRefs[i], wszMember );
    }
}

//=========================================================

void EnumModuleRefs( IMetaDataImport * pIMetaData )
{
    HCORENUM    hEnum = 0;
    mdModuleRef rModuleRefs[2048];
    ULONG       cModuleRefs = ARRAYSIZE(rModuleRefs);
    
    HRESULT hr = pIMetaData->EnumModuleRefs( &hEnum, rModuleRefs, cModuleRefs, &cModuleRefs );
    if ( FAILED(hr) )
        return;

    pIMetaData->CloseEnum( hEnum ); // We don't need the HCORENUM open anymore

    // Spin through and process all the ModuleRefs
    for ( ULONG i = 0; i < cModuleRefs; i++ )
    {
        wchar_t wszName[512];
        ULONG   cchName = ARRAYSIZE(wszName);

        pIMetaData->GetModuleRefProps(
                        rModuleRefs[i], // [IN] moduleref token.    
                        wszName,        // [OUT] buffer to fill with the moduleref name.    
                        cchName,        // [IN] size of szName in wide characters.  
                        &cchName);      // [OUT] actual count of characters in the name.    

        printf( "%s: %08X %ls\n",
                CMetaDataImportHelper::TokenTypeName(rModuleRefs[i]),
                rModuleRefs[i],
                wszName );
    }
}

void FormatGUID( GUID & guid, char * pszOut )
{
    // It's assumed that a buffer of sufficient size
    // will be passed in.
    sprintf( pszOut, "{%08X-%04X-%04X-%02X%02X-%02X%02X%02X%02X%02X%02X}\n",
            guid.Data1, guid.Data2, guid.Data3, guid.Data4[0], guid.Data4[1],
            guid.Data4[2], guid.Data4[3], guid.Data4[4], guid.Data4[5],
            guid.Data4[6], guid.Data4[7] );
}
